from abc import abstractmethod
from pathlib import Path
from typing import Any

from llama_cpp import Llama
import openai

from utils import log_calls


current_directory = Path(__file__).resolve().parent


class Llama2:
    """Loads and queries a Llama2 model."""

    def __init__(
        self,
        model_path: str,
        chat: bool = False,
        n_gpu_layers: int = 60,
        n_threads: int = 8,
        n_ctx: int = 4096,
        verbose: bool = False,
        **kwargs
    ) -> None:
        super().__init__()

        self.chat = chat

        # initialize model
        self.model = Llama(
            model_path,
            chat_format="llama-2",
            verbose=verbose,
            n_gpu_layers=n_gpu_layers,
            n_threads=n_threads,
            n_ctx=n_ctx,
            **kwargs,
        )

    # @log_calls("Running Llama model")
    def __call__(
        self,
        prompt: str,
        chat: bool = None,
        stop: str = "</prompt>",
        max_tokens: int = 200,
        **kwargs: Any
    ) -> str:
        if chat is None:
            chat = self.chat

        if chat:
            return self.model.create_chat_completion(
                # TODO add system message?
                messages=[
                    {
                        "role": "user",
                        "content": prompt,
                    }
                ],
                stop=stop,
                max_tokens=max_tokens,
                **kwargs,
            )["choices"][0]["message"]["content"]
        else:
            return self.model.create_completion(
                prompt=prompt, stop=stop, max_tokens=max_tokens, **kwargs
            )["choices"][0]["text"]


class OpenAI:
    """Queries an OpenAI model using its API."""

    def __init__(
        self, model: str, chat: bool = False, verbose: bool = False, **kwargs
    ) -> None:
        super().__init__()

        self.chat = chat
        self.model = model

        # initialize model
        self.openai_client = openai.OpenAI(**kwargs)

    def __call__(
        self,
        prompt: str,
        chat: bool = None,
        stop: str = "</prompt>",
        max_tokens: int = 200,
        **kwargs: Any
    ) -> str:
        if chat is None:
            chat = self.chat

        if chat:
            return (
                self.openai_client.chat.completions.create(
                    model=self.model,
                    messages=[
                        {
                            "role": "user",
                            "content": prompt,
                        }
                    ],
                    stop=stop,
                    max_tokens=max_tokens,
                    **kwargs,
                )
                .choices[0]
                .message.content
            )
        else:
            return (
                self.openai_client.completions.create(
                    model=self.model,
                    prompt=prompt,
                    stop=stop,
                    max_tokens=max_tokens,
                    **kwargs,
                )
                .choices[0]
                .message.content
            )
