from abc import abstractmethod
from collections import defaultdict
from typing import DefaultDict, Union

from datasets import Dataset, load_dataset
from evaluate import load as load_metric
from tqdm import tqdm
from models import Llama2, OpenAI
from llama_cpp import LlamaGrammar

from utils import log_calls, logger


CLASSIFICATION_PROMPT = """
Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

### Instruction:
{instruction}.

### Input:
{input}

### Response:
"""


class Task:
    def __init__(
        self,
        model: Union[Llama2, OpenAI],
        validation_dataset: str,
        test_dataset: str,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        self.model = model

        self.validation_dataset = load_dataset(
            validation_dataset, split=validation_split
        )
        self.test_dataset = load_dataset(test_dataset, split=test_split)

    @abstractmethod
    def _evaluate(self, prompt: str, dataset):
        pass

    @log_calls("Evaluating validation dataset")
    def evaluate_validation(self, prompt: str):
        return self._evaluate(prompt, self.validation_dataset)

    @log_calls("Evaluating test dataset")
    def evaluate_test(self, prompt: str):
        return self._evaluate(prompt, self.test_dataset)

    @property
    @abstractmethod
    def metric_name(self):
        pass

    @property
    @abstractmethod
    def base_prompt(self):
        pass


# a simple grammar in GBNF notation which allows either positive or negative
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
GRAMMAR = LlamaGrammar.from_string("root ::= (\"positive\"|\"negative\")")

class SentimentAnalysis(Task):
    def __init__(
        self,
        model,
        validation_dataset: str,
        test_dataset: str,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        super().__init__(
            model, validation_dataset, test_dataset, validation_split, test_split
        )

        # whether we use the grammar to constrain the model output or not
        self.use_grammar = True

    def predict(self, prompt: str, text: str):
        # run model for inference using grammar to constrain output
        # TODO grammar also depends on prompt and vice-versa -> what are good labels?
        response = self.model(
            prompt=CLASSIFICATION_PROMPT.format(instruction=prompt, input=text),
            grammar=GRAMMAR
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()
        # input(f"*** PREDICTION***\n\tText: {text}\n\tSentiment: {response}")

        return response

    def _evaluate(self, prompt: str, dataset: Dataset):
        sst2_labels = {"negative": 0, "positive": 1}

        results: DefaultDict[str, int] = defaultdict(int)
        dataset_iterator = tqdm(dataset, desc="evaluating prompt", leave=False)

        for datum in dataset_iterator:
            response = self.predict(prompt=prompt, text=datum["text"]).lower()

            if self.use_grammar:
                # model output is from label space
                answer_label = sst2_labels[response]
            else:
                answer_label = None
                for label in sst2_labels.keys():
                    if label in response:
                        answer_label = sst2_labels[label]
                        break
                else:
                    logger.warning(f"Invalid answer: {response}")
                    results["failed"] += 1
                    continue

            classification_result = (
                "incorrect" if answer_label != datum["label"] else "correct"
            )
            results[classification_result] += 1
            dataset_iterator.set_postfix(results)

        accuracy = results["correct"] / sum(results.values())
        return accuracy

    @property
    def metric_name(self):
        return "accuracy"

    @property
    def base_prompt(self):
        #  from the paper: RLPROMPT: Optimizing Discrete Text Prompts with Reinforcement Learning
        return """In this task, you are given sentences from movie reviews. The task is to classify a sentence as "’positive’" if the sentiment of the sentence is positive or as "’negative’" if the sentiment of the sentence is negative. Return label only without any other text."""


class QuestionAnswering(Task):
    def __init__(
        self,
        model,
        validation_dataset: str,
        test_dataset: str,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        self.evaluation_prompt = """
        Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

        ### Instruction:
        {instruction}

        ### Context:
        {context}

        ### Question:
        {question}

        ### Response:
        """

        self.metric = load_metric("squad")

        super().__init__(
            model, validation_dataset, test_dataset, validation_split, test_split
        )

    def predict(self, prompt: str, context: str, question: str):
        # run model for inference
        # TODO add context-sensitive grammar
        response = self.model(
            prompt=self.evaluation_prompt.format(
                instruction=prompt,
                context=context,
                question=question,
            )
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()

        return response

    def _evaluate(self, prompt: str, dataset: Dataset):
        dataset_iterator = tqdm(dataset, desc="evaluating prompt", leave=False)

        num_samples = 0
        f1 = 0.0
        em = 0
        for datum in dataset_iterator:
            answer = self.predict(
                prompt, context=datum["context"], question=datum["question"]
            ).lower()

            num_samples += 1
            result = self.metric.compute(
                predictions=[{"prediction_text": answer, "id": datum["id"]}],
                references=[{"answers": datum["answers"], "id": datum["id"]}],
            )
            f1 += result["f1"]
            em += result["exact_match"]

            dataset_iterator.set_postfix(
                {"f1": f1 / num_samples, "em": em / num_samples}
            )

        return f1 / num_samples  # , em/num_samples

    @property
    def metric_name(self):
        return "f1"

    @property
    def base_prompt(self):
        # TODO find good prompt
        return """In this task, you are given contexts with questions. The task is to answer the question given the context. Return only the answer without any other text. Make sure that the answer is taken directly from the context."""
