from dataclasses import dataclass, field
import inspect
import json
import logging
from datetime import datetime
from functools import wraps
from hashlib import shake_256
from pathlib import Path
from pprint import pformat
from textwrap import dedent, fill, indent
from typing import Any, Callable

from graphviz import Digraph

current_directory = Path(__file__).resolve().parent
run_directory = (
    current_directory / f"runs/run-{datetime.now().strftime('%Y-%m-%d %H-%M-%S')}"
)
run_directory.mkdir(parents=True, exist_ok=False)
logger = logging.getLogger("test-classifier")
logger.setLevel(level=logging.DEBUG)
file_handler = logging.FileHandler(run_directory / "output.log")
file_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
logger.addHandler(file_handler)


class log_calls:
    description: str

    prolog = dedent(
        """
    ----------------------------------------
    {description}
    {func_name}:
        \tArguments:
        {arguments}
        """
    )

    epilog = dedent(
        """\tResult:
        {result}
    ----------------------------------------
    """
    )

    def __init__(self, description: str, level: int = logging.DEBUG):
        self.description = description
        self.level = level

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            arguments = self._get_named_arguments(func, *args, **kwargs)
            logger.log(
                self.level,
                self.prolog.format(
                    description=self.description,
                    func_name=func.__name__,
                    arguments=indent(pformat(arguments), "\t"),
                ),
            )

            result = func(*args, **kwargs)

            logger.log(
                self.level,
                self.epilog.format(
                    result=indent(pformat(result), "\t"),
                ),
            )
            return result

        return wrapper

    def _get_named_arguments(self, func: Callable[..., Any], *args, **kwargs):
        signature = inspect.signature(func)
        arguments = {}
        for argument_index, (argument_name, argument) in enumerate(
            signature.parameters.items()
        ):
            if argument_index < len(args):
                # argument_name is from args
                value = args[argument_index]
            elif argument_name in kwargs:
                # argument_name is from kwargs
                value = kwargs[argument_name]
            else:
                # argument_name is from defaults
                value = argument.default

            arguments[argument_name] = value
        return arguments


@dataclass(frozen=True)
class Prompt:
    content: str
    score: float
    gen: int
    meta: dict = field(default_factory=dict)

    def __str__(self) -> str:
        return self.content

    def __hash__(self) -> int:
        return (
            hash(self.content)
            + hash(self.score)
            + hash(self.gen)
            + hash(frozenset(self.meta.items()))
        )


class PromptEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Prompt):
            return obj.__dict__
        return json.JSONEncoder.default(self, obj)


@log_calls("Saving genealogy")
def save_genealogy(family_tree, P, metric_name: str):
    node_id = lambda prompt: hash(prompt)
    dot = Digraph(comment="Genealogy")
    for t in range(len(P)):
        with dot.subgraph(name=f"cluster_{t}") as cluster:
            cluster.attr(
                label=f"Generation {t}",
                color="lightgrey",
            )
            for prompt, parents in family_tree:
                if prompt.gen != t:
                    continue
                cluster.node(
                    f"{node_id(prompt)}",
                    label=f"{fill(str(prompt))}\n{metric_name}: {prompt.score}",
                )
                if parents is not None:
                    parent_1, parent_2 = parents
                    dot.edge(f"{node_id(parent_1)}", f"{node_id(prompt)}")
                    dot.edge(f"{node_id(parent_2)}", f"{node_id(prompt)}")

    dot.render(run_directory / "genealogy", engine="dot")


def save_family_tree_visualization(family_tree, P, metric_name: str):
    dot = Digraph(comment="Genealogy")
    node_id = lambda prompt: hash(prompt)
    prompt_generations = lambda prompt: [
        str(t) for t in range(len(P)) if prompt in P[t]
    ]
    node_template = dedent(
        """
        {prompt}
        ∈ P{{{prompt_generations}}}
        {metric_name}: {score}"""
    )
    for prompt, parents in family_tree:
        node_text = node_template.format(
            prompt=fill(str(prompt)),
            prompt_generations=", ".join(prompt_generations(prompt)),
            score=prompt.score,
            metric_name=metric_name
        )
        dot.node(
            f"{node_id(prompt)}",
            label=node_text,
        )
    for prompt, parents in family_tree:
        if parents is not None:
            # TODO consider more than 2 parents?
            dot.edge(
                f"{node_id(parents[0])}",
                f"{node_id(prompt)}",
            )
            dot.edge(
                f"{node_id(parents[1])}",
                f"{node_id(prompt)}",
            )

    dot.render(run_directory / "family_tree", engine="dot")


def save_snapshot(family_tree, P, S, T, N):
    import json

    with open(run_directory / "snapshot.json", "w") as f:
        json.dump(
            {
                "family_tree": family_tree,
                "P": P,
                "S": S,
                "T": T,
                "N": N,
            },
            f,
            indent=4,
            cls=PromptEncoder,
        )


def load_snapshot(path: Path):
    import json

    with path.open("r") as f:
        snapshot = json.load(f)
    return (
        snapshot["family_tree"],
        snapshot["P"],
        snapshot["S"],
        snapshot["T"],
        snapshot["N"],
    )
